( * 4x4 matrix keypad support for 3-instruction Forth - Assad Ebrahim - 2020-07-25---Sat---21:06 )

8 constant  CALL_KEYPAD     \ poll keypad (in 3iKeypad.fs OR 3iDrumPad.fs)

( * 4x4 Matrix Keypad - Membrane or Button - 8 pins * )
( * How it works: )
(     4 pins are tx/output, 4 are rx/input.  )
(     To figure out if/which key was pressed, uC flips one tx pin live at a time, and polls the rx pins )
(     A pressed button will mean we have row/col and logic determines which key was pressed )

\ for gel keypad: TX lines, white tabbed, first four counting from left with pins facing up
\ for pushbutton pcb keypad with yellow pin header: interrogation is left to right col, bottom to top row (S1 to S4, ... S13 to S16)
\ doesn't matter which 4 are input and which 4 are output --- it's symmetric
8  constant padtx1 \ 
7  constant padtx2 
6  constant padtx3 
5  constant padtx4
\ RX lines, black tabbed, last four counting from the left with pins facing up
9  constant padrx1
10  constant padrx2
11 constant padrx3
12  constant padrx4

\ Application: Can map the keys to anything - output or instruction set - e.g. a macrokeypad via Forth...
\ Design: cannot use 74165 shift-in register IC to reduce pin count because only 4 pins are input, other 4 are output

: init-keypad  
	padtx1 out padtx2 out padtx3 out padtx4 out
	padrx1 pullup  padrx2 pullup  padrx3 pullup  padrx4 pullup
;

\ ========== SLOW MODE (polling from host over serial, requires only 3iForth.fs firmware)

\ TODO: what is the timing, with transmit delays?
: poll-keypad ( -- ) \ one poll cycle = 24 serial transmit/responses - c. 1.4s @ 60ms each
	padtx1 low
	padrx1 pinread 1 - if ." S1 " then  
	padrx2 pinread 1 - if ." S2 " then  
	padrx3 pinread 1 - if ." S3 " then  
	padrx4 pinread 1 - if ." S4 " then
	padtx1 high
	
	padtx2 low
	padrx1 pinread 1 - if ." S5 " then  
	padrx2 pinread 1 - if ." S6 " then  
	padrx3 pinread 1 - if ." S7 " then  
	padrx4 pinread 1 - if ." S8 " then
	padtx2 high
	
	padtx3 low
	padrx1 pinread 1 - if ." S9 " then  
	padrx2 pinread 1 - if ." S10 " then  
	padrx3 pinread 1 - if ." S11 " then  
	padrx4 pinread 1 - if ." S12 "   then
	padtx3 high
	
	padtx4 low
	padrx1 pinread 1 - if ." S13 " then  
	padrx2 pinread 1 - if ." S14 " then  
	padrx3 pinread 1 - if ." S15 " then  
	padrx4 pinread 1 - if ." S16 "  then	
	padtx4 high
;

: slow-keypad-poll begin poll-keypad key? until ;


\ ======== FASTMODE - keypad reading put into target uC - requires firmware 3iForth.avr + 3iKeypad.avr
create keypad-phone '1' c, '2' c, '3' c, 'A' c, '4' c, '5' c, '6' c, 'B' c, '7' c, '8' c, '9' c, 'C' c, '*' c, '0' c, '#' c, 'D' c,
create keypad-nums  0 c, 1 c, 2 c, 3 c, 4 c, 5 c, 6 c, 7 c, 8 c, 9 c, 10 c, 11 c, 12 c, 13 c, 14 c, 15 c,

\ DESIGN: key press flag signals through the layers of the loop when to stop polling
variable pollstate
: pollstate! ( n -- ) pollstate ! ;
: pollstate@ ( -- n ) pollstate @ ;
: pollstop  1 pollstate! ;
: resetpoll 0 pollstate! ;

resetpoll

\ Design: read instantaneous value
: keypadread ( -- val )  \ val=0-15 or FF is no key pressed
	CALL_KEYPAD xcall
	100 ms  \ 50ms works but you get multiple reports per press
	1 Nread-com if rx-buffer c@ else $FF then ;

\ Design: keep reading till one key is pressed
\ Design: software polling --- could do other things to interrupt/do other instructions
: fastkeypad ( -- key )  \ key=0-15  (keeps polling if FF no key pressed response)
    begin 
        keypadread dup $FF = if drop 0 else 1 then 
        key? if pollstop then pollstate@ or 
    until ;

\ Design: read till one key press, display to host screen and put on host stack
: fastkeypad-num ( -- keynum )  \ emits one number onto stack
	 cr resetpoll fastkeypad pollstate@ if -1 else bytes keypad-nums + c@ dup . space then ;

\ Design: CONTINUOUS DATA ENTRY until host key is pressed
\ Design: emits numbers pressed onto stack and display
: fastkeypad-nums ( -- <nums> )  
	 cr resetpoll begin fastkeypad pollstate@ if -1 else bytes keypad-nums + c@ dup . space 0 then until ;

\ Design: uses phone translation map 1-9, * 0 #, nothing to stack
: fastkeypad-phone ( -- )  
	 cr resetpoll begin fastkeypad pollstate@ if -1 else bytes keypad-phone + c@ emit space 0 then until ;


\ TODO: implement old-style phone letter selection... if second press within timeout, switch the letter, otherwise accept


\ ====== APPLICATION  (include 3iBeats.fs)
\  KEYPAD TEST -- include 3iKeypad.fs
 \  : beep ( -- ) buzzer hi 5 ms buzzer lo 50 ms ;
 \  : buzzcount ( N -- ) 0 do beep loop ;
 \  : keypad-demo ( -- ) 
 \      begin  fastkeypad 1 + buzzcount  key?  until key drop ;
        
\ APPLICATION - use 74165 to allow a simple ATTINY to monitor a keypad continuously and send out the results, so not to slow down the main uC.  Another application for the GA144...

 \  create beatspad 0 c, 1 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 
 \  : drawnote ( 1/0 -- ) dup 1 = if 14 else 22 then emit space ;
 \  : nibblespace? ( -- ) 1 + 4 mod 0 = if 2 spaces then ;  \ put space after each 4 bits (nibble) to aid visual perception
 \  : beats-demo ( -- )
 \  cr ." Enter a quartet of 16th-note beats using hex digits on keypad: " cr
 \      4 0 do fastkeypad bytes beatspad + c@ drawnote i nibblespace? loop >b4 1 Nseq-nibble-msb  \ get 8 entries, then loop it
 \  ; 

\ ======== KEYPAD PORTION =========
\ init-keypad
\  : beat-keypad ( -- )  \ input rhythm from 4x4 matrix keypad, 1 quarter note (4 beats) at a time.
\                     \ ex: 8 8 11 0   10 11 0 0   8 8 11 0   10 15 8 0
\      reset-stop resetbytes draw-menu
\      cr ." Input beats (hex digits)." cr
\      begin fastkeypad-num bytecount++ bytecount@ nseq-nibble-msb key? until key drop
\  ;
\ TODO: wrap into menu...
\ _________________________________


\ ====== APPLICATION \ ======== 5 button menu
( BUTTON LAYOUT:
    k3-play k2-clr k1-rest
     k4-clrall k5-beat
)

\ 5-button menu controls use Nano pins 2-6
\ 2 constant k1-rest
\ 3 constant k2-clr
\ 4 constant k3-play
\ 5 constant k4-clrall
\ 6 constant k5-beat
\ 
\ : buttons-init
\     k1-rest pullup \ buttons connect pin to ground and are active low
\     k2-clr pullup
\     k3-play pullup
\     k4-clrall pullup
\     k5-beat pullup
\ ;




\ =================================================

: help-keypad
  cr
  cr  ." KEYPAD (help-keypad)"
  cr 
  cr  ."  Firmware: 3iKeypad.avr (NESTED2) 3iI2C.avr  3iForth.avr"
  cr  ." include 3iForth.fs  3iKeypad.fs"
  cr  ." 1 open-com   rx"
  cr
  cr  ." 4X4 MATRIX KEYPAD"
  cr  ." init-keypad   create keypad (16 values)"
  cr  ." slow-keypad-poll (requires only 3iForth.fs)"
  cr  ." fastkeypad-phone   fastkeypad-nums"
  cr
;

help-keypad


\ Version History
\ V1 - 2020-07-25---Sat---21:06
\ V2 - 2021-09-30---Thu---07:46 - fixed and remapped key pin order for linear decode mappings
