\ 3iMotor.fs - 2020-10-11---Sun---14:51


\ Hardware:
\ 28BYJ-48 5V DC Step Motor, 4 coils, with 3 ways to drive: 
\       wave (A B C D) , full (AB BC CD AD ), or half (A AB B BC C CD D AD)
\       5.625 rotor half-step angle, or 64 steps per revolution
\       The motor has an internal gearbox with 1:63.68395 additional gear reduction => max 4076 microsteps per revolution in half step mode
\           => 0.08833* per micro halfstep  => 12 micro halfsteps per degree (6 steps in wave mode)

\ ULN2003A-PCB motor driver PCB

\ Connections:
\ ports 5 4 3 2 out from Nano to ULN board (2 to IN1)
\ motor to ULN board --- blue to top (aligned with IN1), red at bottom (power)
\ WARNING: power the ULN board separately with 5V from DC volt
\ jumper pin powers the board or cuts power to it -- use a relay to cut power to the motors (they will draw power ambiently)

\ Dec 25, 2018 - 9V to 5V DC converter producing 5V TTL output to 4-button manual step motor driver

: stepper-init  5 4 3 2 port4out ;
: test-port4  16 0 do i port4write 250 ms loop 0 port4write ;

19 constant REG_MOTORCODE

$17 constant CALL_MOTORDEG


: deg>> ( deg delayms -- )   \ Move motor deg degrees in relative measure, with ms delay between steps (2 is smallest)
    xstore 
    dup 0< if 0 xstore negate else 1 xstore then  \ change neg flag value to reverse direction based on motor mounting
    lohi swap xstore xstore REG_PORTAOUT xstore REG_MOTORCODE xstore 
    CALL_MOTORDEG xcall ;
\ ACTION -- can check if motor is done by reading for serial
\ FEATURE -- start by flushing rx queue

\ Speed: top speed is based on gearbox, controller, and driver code (drive mode and delay).  Note: faster means less torque.
: ## ( deg -- ) 2 deg>>  ;  \ Fastest delay is 2 MS (1 rev in 4 sec or 15 rpm).  Less than this and motor skips even with no load.
: ### ( deg -- ) 4 deg>> ;  \ Delay of 4 MS (8 rpm) allows quick changes in rotation direction with no load, e.g. aim, fire, re-aim
: #### ( deg -- ) 1 deg>> ; 

: test-steppers  110 ## -60 ## 40 ## -130 ## 40 ## ;  \ NOTE - these are asynchronous commands, i.e. there is no waiting for rotation to finish    

\ ============== USING 3iMotor primitives deg()


\ ========================   USING 3iForth Primitives only, and building up Motor movements on remote Forth (less efficient) ====================

\ Design A: Use a variable that holds where in the sequence we are, whether stepping forward or backwards.  This means no jitters or loss of position.
\ This is preferred over Design C: c@+, c@-, over Design B: lshift / rshifts (actually should be rotates)

\ Motor drive methods.  
create wave4 1 c, 2 c, 4 c, 8 c,   \ single coil, half power, half torque, 32 steps per rev, 160mA current draw at 5V = 800mW
create full4 3 c, 6 c, 12 c, 9 c,  \ double coil, full power, full torque, 32 steps per rev, 320mA current draw at 5V = 1.6W
create half8 1 c, 3 c, 2 c, 6 c, 4 c, 12 c, 8 c, 9 c,  \ alternating coils, power, and torque, 64 steps per rev 
\ half8 = 4096 steps per revolution (2^12)

\ WAVE MODE, single phase, lowest current, lowest torque
variable mm  \ hold where in the motor code sequence we are, whether stepping forwar or backwards.  Means no jitters or loss of position.
: mm! ( m -- ) mm c! ;  \ store
: mm@ ( -- m ) mm c@ ;  \ fetch
1 mm! \ initialize
: >m  ( idx -- ) wave4 + c@ port4write ;   \ send command to motor
: mm+ ( -- m ) mm@ 1 + 4 mod dup mm! >m ;  \ STEP (rotate) through motor codes, sending instruction to motor
: mm- ( -- m ) mm@ 1 - 4 mod dup mm! >m ;
: spinN ( N -- ) \ 2048 steps per revolution (2^11)  \ spin through steps
    dup 0<> if 
        dup 0> if 0 do mm+ loop else negate 0 do mm- loop then 
    else drop then ;  
28 constant 5degsector  \ 28 steps in 5 deg sector (sectors: 3 (15*), 9 (45*), 18 (90*), 37 (180), 73 (360)
: sectors ( N -- ) 5degsector * spinN ;
: deg ( N -- ) dup 5 * swap 31 * 45 / + spinN ;   \  2048 steps per revolution => 5.6888 steps per degree, or 5 and 2/3 (2040 steps/rev, 0.39% error) or 31/45


\ ==============  SERVO MOTORS are used to signal specific encoded positions and the motor runs to those
\ They have a built in PCB that interprets the PWM signals and encodes the shaft etc.

3 constant laser

create servos 3 bytes allot   \ holds which pins is mapped to servo #
: set-servo ( pin servo -- ) servos + c! ;
: servo ( servo -- pin ) servos + c@ ;
: servo-init ( -- )  \ SG09 servo, 180* span, 50Hz/20ms envelop  --- servo# is 0,1,2,  only use pins 9,10
    10 0 set-servo  \ PAN  / YAW
    9  1 set-servo  \ TILT / PITCH
    11 2 set-servo  \        ROLL
    4 1 setTimer 2000 ms  \ 122 Hz, 256 prescale,  16 31 47 62 78 counts,  affects pins 9/10 only
    6 2 setTimer 2000 ms  \ 122 Hz, 256 prescale,  16 31 47 62 78 counts,  affects pins 3/11 only
    laser out
    ." Done." ;
    
: servo-close
    3 0 do i servo dup out low 200 ms loop 
    3 1 setTimer 2000 ms
    4 2 setTimer 2000 ms ; \ 490 Hz, 64 prescale,  62 124 187 249 x counts -- DEFAULT
    
\ ATTEMPT AT CAPTURING THE MATH
\ : f490 1 0 d>f 490 0 d>f f/ ;  \ 490 Hz -> wavelength
\ : s490 f490 255 0 d>f f/ ;     \ 490 Hz -> steplength
\ 
\ : f980 1 0 d>f 980 0 d>f f/ ;
\ : f50  1 0 d>f 50  0 d>f f/ ;

variable myservo \ hold which servo is current
: myservo! ( servo# -- ) myservo ! ;
: myservo@ ( --  servo# ) myservo @ ;

: gopos ( pwm -- ) myservo@ servo pwm ;
: gopos_ ( pwm dly -- ) >r gopos r> ms ;
: goseq ( ... n ) dup 0 do dup i - roll 500 gopos_ loop drop ;  \ picks the position stack from the beginning

\ default timer  490 Hz
\ : testservo1 48 gopos 500 ms 254 gopos 500 ms 151 gopos 500 ms ;
\ : testservo2 48 500 gopos_ 254 500 gopos_ 151 500 gopos_ ;
\ : testservo3 48 254 48 254 48 254 151 48 151 254 151 servo goseq ;
: testservo490 62 124 187 249 187 124 62 187 62 187 10 goseq ;  \ 0 45 90 135

: demoservo 16 31 47 62 78 62 47 31 16 31 47 16 62 16 78 47    16  goseq  ;  \ 122 Hz timer 1  pins 9,10

: cone90 47 31 47 62 47    5 goseq ;  \ move within 90 deg cone 
: cone180 16 31 47 62 78 62 47 31 16 47   10 goseq ;

: pan-servo 0 myservo! ;
: tilt-servo 1 myservo! ;
: roll-servo 2 myservo! ;
: recenter 47 gopos ;
: servo-min 16 gopos ;
: servo-max 78 gopos ;
: servo-mid 47 gopos ;
: servo-q1 31 gopos ;
: servo-q3 62 gopos ;


: fire 1000 ms laser hi 1000 ms laser lo ;
: pan-fire pan-servo 47 gopos fire 31 gopos fire 16 gopos fire 62 gopos fire 78 gopos fire 47 gopos fire ;
: tilt-fire tilt-servo 47 gopos fire 31 gopos fire 62 gopos fire 47 gopos fire ;
: roll-fire roll-servo 47 gopos fire 31 gopos fire 62 gopos fire 47 gopos fire ;

: reset-laser pan-servo recenter tilt-servo recenter roll-servo recenter pan-servo ;

: demo-3dof-robotic-laser  laser out pan-fire tilt-fire roll-fire  reset-laser ;

: demo-cool-3dof  
    laser out reset-laser fire 
    pan-servo servo-max fire 
    tilt-servo servo-max fire
    roll-servo servo-max fire
    pan-servo servo-min fire
    roll-servo servo-min fire
    tilt-servo servo-min fire
    reset-laser fire  ;
    

: help-motors
cr
cr ." HELP-MOTORS"
cr ." -----------"
cr ."   stepper-init"
cr ."   test-port4  - lights up 4-led output board 0x0 to 0xF"
cr ."   5 sectors   - moves 5x 5-sector degrees using host 3iForth primitives " 
cr ."   2 25 deg>>  - moves 25 deg with 2 ms delay between steps (1 is fastest)"
cr ."   25 ##       - 2ms delay - moves 25 degrees forward using remote Motor primitives"
cr ."   -25 ##      - 2ms delay - moves 25 degrees backward using remote Motor primitives"
cr ."   25 ###      - 4ms delay - moves 25 degrees forward using remote Motor primitives"
cr ."   test-steppers - operates 5 back and forth moves ASYNCHRONOUSLY (ie without feedback loop/servo) at speed 2"
cr
cr ."   servo-init ( pin -- ) - use pins 10, 9, 11 for pan, tilt, roll axes"
cr ."   reset-laser"
cr ."   pan-servo  tilt-servo  roll-servo  -- selects the axis from 3DOF"
cr ."   gopos   16 31 47 62 78  are the 45 degree points on the 180 span"
cr ."   cone90  cone180  servo-min  servo-q1  servo-mid  servo-q3  servo-max -- movement choices"
cr ."   fire - shoot laser"
cr ."   demo-3dof-robotic-laser   demo-cool-3dof"
cr ."   servo-close "
cr ;

\ =====================================================
help-motors


\ NEXT STEP:
\ Why does motor get hot?  How to cool it?
\ To go faster, incorporate the spinN code in the AVR
\ Add relay to cut power
\ Measure power consumption
\ Do I really need the ULN PCB, why not just the chip?
\ What about TIP122 transistor, would I need 4 therefore?
\ Why 512 steps in a revolution?  
\ Get a continuous rotate left / rotate right operator...

\ Add HOME and SETPOS functionality
\ Push button and spin fixed amount
\ Press button and keep spinning until release or until second/different press
\ Joystick control
\ Speed control


\ Versions:
\ V1 - 2020-10-11---Sun---14:51 - Added stepper motor control 28BYJ-48
\ V2 - 2020-11-08---Sun---14:08 - Added servo motor control SG90
\    - 2020-11-08---Sun---21:28 - Made servo usage easy with 3DOF pan,tilt,roll (yaw, pitch, roll)
\    - 2020-11-08---Sun---23:25 - STREAMLINED/FEATURE: concluded demo for Servo Motors (Timor the Terrible Battle Robot)
